#include <math.h>
#include <string.h>
#include <stdio.h>
#include <SDL.h>
#include <SDL_audio.h>

#include "Constants.h"
#include "Timing.h"
#include "Log.h"
#include "SoundBuffer.h"

//const int SAMPLE_RATE_PER_SECOND = 320000;
const int SAMPLE_RATE_PER_SECOND = 44100;
//const int SAMPLE_RATE_PER_SECOND = 22050;
//const int SAMPLE_RATE_PER_SECOND = 11025;
//const int SAMPLE_RATE_PER_SECOND = 8000;

const Uint8 NUMBER_OF_CHANNELS = 2; // stereo

const Sint16 AUDIO_HIGH_I = 0x3DF0;
const Sint16 AUDIO_LOW_I = 0x0000;

const float AUDIO_HIGH_F = 0.10f;
const float AUDIO_LOW_F = 0.0f;

static Uint8 _soundInitialized = 0;

//    ++-----------------------sample is signed if set
//    ||
//    ||       ++-----------sample is bigendian if set
//    ||       ||
//    ||       ||          ++---sample is float if set
//    ||       ||          ||
//    ||       ||          ||     +---sample bit size---+
//    ||       ||          ||     |                     |
//    15 14 13 12 11 10 09 08     07 06 05 04 03 02 01 00
SDL_AudioFormat _format;
Uint8 _formatBitsPerSample;
Uint8 _formatIsFloat;

// used to decide when to sample the speaker, writing into the sound buffer
static Uint32 _spectrumSpeakerSamplingIntervalTstates;
static Uint32 _tstatesAtLastSample;

// the write operation treats the buffer as linear, to not overwrite earlier samples
void _sound_spectrum_buffer_write(Uint8 value8) {
    if (!_soundInitialized) {
        return;
    }

    soundbuffer_write(value8);
    return;
}

Uint8 _sound_spectrum_buffer_read() {
    if (!_soundInitialized) {
        return 0xF3;
    }

    return soundbuffer_read();
}

void sound_play() {
    if (!_soundInitialized) {
        return;
    }

    SDL_PauseAudio(0);      // start playing sound
}

void _sound_buffer_filler_callback(void* consumerData, Uint8* sdlBuffer, int bytesToFill) {
    int lengthInSamples = bytesToFill / (_formatBitsPerSample/8);
    lengthInSamples = lengthInSamples / NUMBER_OF_CHANNELS;

    if (_formatIsFloat) {
        // samples must be floating point

        if (_formatBitsPerSample == 32) {
            float* buffer = (float*)sdlBuffer;
            for (int i = 0; i < lengthInSamples*NUMBER_OF_CHANNELS; i+=NUMBER_OF_CHANNELS) {
                float value = _sound_spectrum_buffer_read() ? AUDIO_HIGH_F : AUDIO_LOW_F;
                //if (value == AUDIO_HIGH_F)log_write("1"); else log_write("0");
                for (int j = 0; j < NUMBER_OF_CHANNELS; j++) {
                    buffer[i + j] = value;
                }
            }
        }
        else if (_formatBitsPerSample == 64) {
            double* buffer = (double*)sdlBuffer;
            for (int i = 0; i < lengthInSamples * NUMBER_OF_CHANNELS; i += NUMBER_OF_CHANNELS) {
                double value = _sound_spectrum_buffer_read() ? (double)AUDIO_HIGH_F : (double)AUDIO_LOW_F;
                for (int j = 0; j < NUMBER_OF_CHANNELS; j++) {
                    buffer[i + j] = value;
                }
            }
        }
    }
    else {
        // samples must be integers

        if (_formatBitsPerSample == 16) {
            Uint16* buffer = (Uint16*)sdlBuffer;
            for (int i = 0; i < lengthInSamples * NUMBER_OF_CHANNELS; i += NUMBER_OF_CHANNELS) {
                Uint16 value = _sound_spectrum_buffer_read() ? (Uint16)AUDIO_HIGH_I : (Uint16)AUDIO_LOW_I;
                for (int j = 0; j < NUMBER_OF_CHANNELS; j++) {
                    buffer[i + j] = value;
                }
            }
        }
        else if (_formatBitsPerSample == 32) {
            Uint32* buffer = (Uint32*)sdlBuffer;
            for (int i = 0; i < lengthInSamples * NUMBER_OF_CHANNELS; i += NUMBER_OF_CHANNELS) {
                Uint32 value = _sound_spectrum_buffer_read() ? (Uint32)AUDIO_HIGH_I : (Uint32)AUDIO_LOW_I;
                for (int j = 0; j < NUMBER_OF_CHANNELS; j++) {
                    buffer[i + j] = value;
                }
            }
        }
    }
}

// called after each CPU instruction, it is in charge of deciding
// if it's time to write to the sound buffer
void sound_add_sample(Uint8 isSpeakerActive, Uint64 totalTStates) {
    if (!_soundInitialized) {
        return;
    }

    if (totalTStates - _tstatesAtLastSample >= _spectrumSpeakerSamplingIntervalTstates) {
        // this allows for a catchup on consecutive CPU instructions, if, for some reason,
        // we fall behind for more than one sampling interval
        _tstatesAtLastSample += _spectrumSpeakerSamplingIntervalTstates;
        _sound_spectrum_buffer_write(isSpeakerActive);
    }
}

void sound_notify_end_of_frame() {
    if (!_soundInitialized) {
        return;
    }
}

void sound_destroy() {
    if (!_soundInitialized) {
        return;
    }

    _soundInitialized = 0;

    SDL_PauseAudio(1); // stop playing sound
    SDL_CloseAudio();

    soundbuffer_destroy();
}

void sound_start() {
    sound_destroy();

    if (SDL_Init(SDL_INIT_AUDIO) != 0) {
        log_write("Error: could not initialize SDL audio");
        log_write((char*)SDL_GetError());
        return;
    }

    SDL_AudioSpec want;
    want.freq = SAMPLE_RATE_PER_SECOND;
    want.format = AUDIO_F32SYS;
    want.channels = NUMBER_OF_CHANNELS;
    want.samples = 0x1000;  // buffer size
    want.callback = _sound_buffer_filler_callback;
    want.userdata = NULL;   // unused
    want.silence = 0;

    char* message = (char*)malloc(500 * sizeof(char));
    if (message) {
        sprintf_s(message, 490, "Requesting audio spec %04x", want.format);
        log_write(message);
        free(message);
    }

    SDL_AudioSpec got;
    if (SDL_OpenAudio(&want, &got) != 0) {
        log_write("Error: could not open audio");
        log_write((char*)SDL_GetError());
        return;
    }
    _format = got.format;
    _formatBitsPerSample = SDL_AUDIO_BITSIZE(got.format);
    _formatIsFloat = SDL_AUDIO_ISFLOAT(got.format) ? 1 : 0;

    message = (char*)malloc(500 * sizeof(char));
    if (message) {
        sprintf_s(message, 490, "Got audio spec %04x", got.format);
        log_write(message);
        free(message);
    }

    if (SDL_AUDIO_ISBIGENDIAN(got.format)) {
        log_write("Warning: got audio spec that is big endian, which is unsupported; audio will likely be silence or garbage");
    }

    if (want.format != got.format || 
        want.channels != got.channels ||
        want.freq != got.freq) {

        log_write("Warning: got an audio spec different than the one requested");
    }

    // calculate ZX Spectrum speaker buffer size... it must be large enough
    // to hold enough samples to last as long as a video frame
    Uint16 soundFrameDurationMs = timing_get_milliseconds_per_video_frame();

    Uint32 _samplesPerFrame = (Uint32)((double)SAMPLE_RATE_PER_SECOND * ((double)soundFrameDurationMs / 1000.0f));
    // calculate sampling interval in tstates
    Uint64 tstatesPerVideoFrame = timing_get_states_per_video_frame();
    _spectrumSpeakerSamplingIntervalTstates = (Uint32)((double)tstatesPerVideoFrame / (double)_samplesPerFrame);

    _tstatesAtLastSample = 0;

    soundbuffer_start(SAMPLE_RATE_PER_SECOND);
    _soundInitialized = 1;
}
