#include <Windows.h>
#include <sys\timeb.h>
#include <SDL.h>
#include <stdio.h>

#include "Constants.h"
#include "Timing.h"
#include "Io.h"
#include "ULABus.h"
#include "Memory.h"
#include "Cpu.h"
#include "Log.h"
#include "Sound.h"
#include "TapUi.h"

static SDL_Surface* _surface;
static SDL_Renderer* _renderer;
static SDL_Window* _window;

static BOOL _videoIsInitialized = FALSE;
SDL_Rect _entireWindowRectangle;

HANDLE _callbackTimerHandle;
HANDLE _videoFrameMutex;

static volatile char _isDestroying;

static volatile char _timerCallbackDestroyed;
static SDL_TimerID _timer;

struct timeb _videoStart, _videoEnd;

static Uint64 _frames = 0;
static Uint64 _framesAtNormalSpeed = 0;
static Uint64 _tstates = 0;
static Uint8 _zoom = 1;
static Uint8 _isFlashInverse;

static Uint8 _palette[] = {
    0x00, 0x00, 0x00,   // black
    0x00, 0x22, 0xc7,   // blue
    0xd6, 0x28, 0x16,   // red
    0xd4, 0x33, 0xc7,   // magenta
    0x00, 0xc5, 0x25,   // green
    0x00, 0xc7, 0xc9,   // cyan
    0xcc, 0xc8, 0x2a,   // yellow
    0xca, 0xca, 0xca,   // white

    // bright versions
    0x00, 0x00, 0x00,   // black
    0x00, 0x2b, 0xfb,   // blue
    0xff, 0x33, 0x1c,   // red
    0xff, 0x40, 0xfc,   // magenta
    0x00, 0xf9, 0x2f,   // green
    0x00, 0xfb, 0xfe,   // cyan
    0xff, 0xfc, 0x36,   // yellow
    0xff, 0xff, 0xff,   // white
};

// ================================ BEGIN CACHE AREA ==================================
//
// I experimented with saving pointers to individual pixel addresses - since, for
// example, at zoom=2, 4 pixels must be written to surface memory for each ZX Spectrum
// pixel.
//
// Very surprisingly, this approach of caching pixel pointers was found to 
// not take noticeably less CPU time every frame, so it remains unused.

// each (col, row) contains an array whose elements are pointers into surface's pixel array
// of all the on-screen pixels needed to draw the ZX Spectrum pixel at (col, row)
// each (col, row) location needs _zoom*_zoom such offsets
// 
// it is used to remove a large number of multiplications from each video frame
//Uint32** _video_pixel_cache[VISIBLE_SPECTRUM_SCREEN_HEIGHT][VISIBLE_SPECTRUM_SCREEN_WIDTH];

//void _video_preload_pixel_cache() {
//    Uint32* p_screen = (Uint32*)_surface->pixels;
//    
//    int xLimit = VISIBLE_SPECTRUM_SCREEN_WIDTH;
//    int yLimit = VISIBLE_SPECTRUM_SCREEN_HEIGHT;
//
//    for (int x = 0; x < VISIBLE_SPECTRUM_SCREEN_WIDTH; x++) {
//        for (int y = 0; y < VISIBLE_SPECTRUM_SCREEN_HEIGHT; y++) {
//
//            int realPixelIndex = 0;
//            _video_pixel_cache[y][x] = (Uint32**)malloc((_zoom * _zoom) * sizeof(Uint32**));
//            if (!_video_pixel_cache[y][x]) {
//                log_write("Error: unable to allocate memory for video pixel lookup");
//                return;
//            }
//
//            for (int yy = 0; yy < _zoom; yy++) {
//                int offset = (_surface->pitch / sizeof(Uint32)) * (y * _zoom + yy) + (x * _zoom);
//                if ((y * _zoom + yy) < 0 || (y * _zoom + yy) >= yLimit * _zoom) {
//                    continue;
//                }
//                for (int xx = 0; xx < _zoom; xx++) {
//                    if ((x * _zoom + xx) < 0 || (x * _zoom + xx) >= xLimit * _zoom) {
//                        continue;
//                    }
//                    _video_pixel_cache[y][x][realPixelIndex++] = (Uint32*)(p_screen + offset++);
//                }
//            }
//        }
//    }
//}

//void _video_set_cached_pixel(int x, int y, Uint8 r, Uint8 g, Uint8 b, Uint8 a) {
//    Uint32 hardwarePixelValue = SDL_MapRGBA(_surface->format, r, g, b, a);
//
//    //x -= HIDE_LEFT_BORDER_COUNT;
//    //y -= HIDE_TOP_BORDER_COUNT;
//
//    if (x /*- HIDE_LEFT_BORDER_COUNT*/ >= VISIBLE_SPECTRUM_SCREEN_WIDTH) {
//        return;
//    }
//
//    if (y /*- HIDE_TOP_BORDER_COUNT*/ >= VISIBLE_SPECTRUM_SCREEN_HEIGHT) {
//        return;
//    }
//
//    for (int c = 0; c < _zoom * _zoom; c++) {
//        *(_video_pixel_cache[y][x][c]) = hardwarePixelValue;
//    }
//}

// ================================ END CACHE AREA ==================================

void _video_run_cpu_for_scanline() {
    Uint32 allocation = timing_get_states_per_scanline();
    Uint64 tstatesConsumed = cpu_run_many(allocation);

    if (!timing_is_faster()) {
        // we only care about statistics when we're running at normal speed
        _tstates += tstatesConsumed;

        // same for more accurate timing
        timing_report_actual_states_last_scanline((Uint32)tstatesConsumed);
    }
}

inline void _video_set_pixel(int x, int y, Uint8 r, Uint8 g, Uint8 b, Uint8 a) {
    Uint32* p_screen = (Uint32*)_surface->pixels;
    Uint32 hardwarePixelValue = SDL_MapRGBA(_surface->format, r, g, b, a);

    x -= HIDE_LEFT_BORDER_COUNT;
    y -= HIDE_TOP_BORDER_COUNT;

    int xLimit = VISIBLE_SPECTRUM_SCREEN_WIDTH;
    int yLimit = VISIBLE_SPECTRUM_SCREEN_HEIGHT;

    for (int yy = 0; yy < _zoom; yy++) {
        int offset = (_surface->pitch / sizeof(Uint32)) * (y*_zoom + yy) + (x*_zoom);
        if ((y*_zoom + yy) < 0 || (y * _zoom + yy) >= yLimit * _zoom) {
            continue;
        }
        for (int xx = 0; xx < _zoom; xx++) {
            if ((x * _zoom + xx) < 0 || (x * _zoom + xx) >= xLimit * _zoom) {
                continue;
            }
            *(p_screen + offset++) = hardwarePixelValue;
        }
    }
}

void _video_write_border_pixel(int x, int y) {
    Uint8 value = io_read8_border_colour() & 0x07;   // keep only colour

    Uint8 red = *(_palette + 3 * value + 0);
    Uint8 green = *(_palette + 3 * value + 1);
    Uint8 blue = *(_palette + 3 * value + 2);
    _video_set_pixel(x, y, red, green, blue, 255);
}

void _video_write_border_scan_line(int y) {
    for (int i = 0; i < SPECTRUM_SCREEN_WIDTH; ++i) {
        _video_write_border_pixel(i, y);
    }
}

void _video_write_paper_scan_line(int y, int rowStartAddress) {
    int x = 0;

    // left border
    for (int i = 0; i < SPECTRUM_BORDER_WIDTH; i++) {
        _video_write_border_pixel(x + i, y);
    }
    x += SPECTRUM_BORDER_WIDTH;

    int attributesBaseAddress = 16384 + (SPECTRUM_PAPER_HEIGHT * SPECTRUM_PAPER_WIDTH) / 8;    // this is now at the start of attributes RAM
    int attributeRowNumber = (y - SPECTRUM_BORDER_LINES_BEFORE_PAPER) / 8;
    int attributeRowOffset = attributeRowNumber * (SPECTRUM_PAPER_WIDTH / 8);

    for (int i = 0; i < SPECTRUM_PAPER_WIDTH; i+=8) {
        Uint8 pixelByte = memory_read8(rowStartAddress + i/8);
        ula_bus_write8(pixelByte);
        
        int attributeByteOffset = attributeRowOffset + i / 8;

        int attributeAddress = attributesBaseAddress + attributeByteOffset;
        Uint8 attributeByte = memory_read8(attributeAddress);
        ula_bus_write8(attributeByte);

        // figure out R, G, B for the next 8 pixels
        // attributes
        // bits:       7        6     5   4   3      2   1   0
        //         FLASH   BRIGHT    P2  P1  P0     I2  I1  I0
        Uint8 isBright = attributeByte & 0x40;
        Uint8 isInverse = (attributeByte & 0x80) && _isFlashInverse;

        Uint8 inkColour = attributeByte & 0x07;
        Uint8 inkColourOffset = 3 * inkColour;
        if (isBright) {
            inkColourOffset = inkColourOffset + 3 * SPECTRUM_COLOUR_COUNT;
        }
        
        Uint8 inkRed = *(_palette + inkColourOffset + 0);
        Uint8 inkGreen = *(_palette + inkColourOffset + 1);
        Uint8 inkBlue = *(_palette + inkColourOffset + 2);

        Uint8 paperColour = (attributeByte >> 3) & 0x07;
        Uint8 paperColourOffset = 3 * paperColour;
        if (isBright) {
            paperColourOffset = paperColourOffset + 3 * SPECTRUM_COLOUR_COUNT;
        }

        Uint8 paperRed = *(_palette + paperColourOffset + 0);
        Uint8 paperGreen = *(_palette + paperColourOffset + 1);
        Uint8 paperBlue = *(_palette + paperColourOffset + 2);

        for (int j = 0; j < 8; j++) {
            if (!isInverse) {
                if (!(pixelByte & 0x80)) {
                    // paper colour
                    _video_set_pixel(x + i + j, y, paperRed, paperGreen, paperBlue, 255);
                }
                else {
                    // ink colour
                    _video_set_pixel(x + i + j, y, inkRed, inkGreen, inkBlue, 255);
                }
            }
            else {
                if (!(pixelByte & 0x80)) {
                    // ink colour
                    _video_set_pixel(x + i + j, y, inkRed, inkGreen, inkBlue, 255);
                }
                else {
                    // paper colour
                    _video_set_pixel(x + i + j, y, paperRed, paperGreen, paperBlue, 255);
                }
            }
            pixelByte = pixelByte << 1;
        }
    }
    x += SPECTRUM_PAPER_WIDTH;

    // right border
    for (int i = 0; i < SPECTRUM_BORDER_WIDTH; i++) {
        _video_write_border_pixel(x + i, y);
    }
}

inline Uint8 _video_get_must_render_pixels() {
    if (!timing_is_faster()) {
        // normal CPU timing, so we render pixels every frame
        return 1;
    }

    // turbo mode is on, so rarely draw pixels
    if ((_frames & 0x02) == 0x02) {
        return 1;
    }

    return 0;
}

void _video__preframe() {
    if (_isDestroying) {
        _timerCallbackDestroyed = 1;
        return;
    }

    _isFlashInverse = _frames & 0x10;
    cpu_request_interrupt();
}

void _video__postframe() {
    if (!timing_is_faster()) {
        _framesAtNormalSpeed++;
    }

    double msPerFrame = (double)timing_get_milliseconds_per_video_frame();
    Uint64 framesPerSecond = (Uint64)(1000.0f / msPerFrame);
    Uint64 soundFrameDelay = framesPerSecond / 10;
    if (_frames == 1) {
        // delay start of sound to accumulate some samples
        sound_play();
    }

    _frames++;
    
    sound_notify_end_of_frame();
}

void _video__frame_with_pixels() {
    SDL_LockSurface(_surface);

    int y = 0;

    // border before (above) paper
    for (int i = 0; i < SPECTRUM_BORDER_LINES_BEFORE_PAPER; i++) {
        _video_write_border_scan_line(y + i);
        _video_run_cpu_for_scanline();
    }
    y += SPECTRUM_BORDER_LINES_BEFORE_PAPER;

    int rowStartAddressAtStartOfThird = 16384;

    // top third of paper
    for (int k = 0; k < SPECTRUM_PAPER_HEIGHT / 3; k++) {
        int mod = k % 8;
        int div = k / 8;
        int rowStartAddress = rowStartAddressAtStartOfThird + (mod * SPECTRUM_PAPER_HEIGHT / 24) * (SPECTRUM_PAPER_WIDTH / 8) + (div * SPECTRUM_PAPER_WIDTH / 8);
        _video_write_paper_scan_line(y + k, rowStartAddress);
        _video_run_cpu_for_scanline();
    }
    y += SPECTRUM_PAPER_HEIGHT / 3;
    rowStartAddressAtStartOfThird += (SPECTRUM_PAPER_WIDTH / 8) * (SPECTRUM_PAPER_HEIGHT / 3);

    // middle third of paper
    for (int k = 0; k < SPECTRUM_PAPER_HEIGHT / 3; k++) {
        int mod = k % 8;
        int div = k / 8;
        int rowStartAddress = rowStartAddressAtStartOfThird + (mod * SPECTRUM_PAPER_HEIGHT / 24)*(SPECTRUM_PAPER_WIDTH / 8) + (div * SPECTRUM_PAPER_WIDTH / 8);
        _video_write_paper_scan_line(y + k, rowStartAddress);
        _video_run_cpu_for_scanline();
    }
    y += SPECTRUM_PAPER_HEIGHT / 3;
    rowStartAddressAtStartOfThird += (SPECTRUM_PAPER_WIDTH / 8) * (SPECTRUM_PAPER_HEIGHT / 3);

    // bottom third of paper
    for (int k = 0; k < SPECTRUM_PAPER_HEIGHT / 3; k++) {
        int mod = k % 8;
        int div = k / 8;
        int rowStartAddress = rowStartAddressAtStartOfThird + (mod * SPECTRUM_PAPER_HEIGHT / 24) * (SPECTRUM_PAPER_WIDTH / 8) + (div * SPECTRUM_PAPER_WIDTH / 8);
        _video_write_paper_scan_line(y + k, rowStartAddress);
        _video_run_cpu_for_scanline();
    }
    y += SPECTRUM_PAPER_HEIGHT / 3;

    ula_bus_go_inactive();

    // border after (below) paper
    for (int i = 0; i < SPECTRUM_BORDER_LINES_AFTER_PAPER; i++) {
        _video_write_border_scan_line(y + i);
        _video_run_cpu_for_scanline();
    }

    SDL_UnlockSurface(_surface);

    // convert to a texture the surface to which we've drawn the pixels
    SDL_Texture* entireWindowTexture = SDL_CreateTextureFromSurface(_renderer, _surface);
    // render play area
    SDL_RenderCopy(_renderer, entireWindowTexture, NULL, &_entireWindowRectangle);
    SDL_DestroyTexture(entireWindowTexture);

    // render TAP UI
    tapui_render(_renderer);

    // all rendering is done
    SDL_RenderPresent(_renderer);
}

void _video__frame_without_pixels() {
    int y = 0;

    // border before (above) paper
    for (int i = 0; i < SPECTRUM_BORDER_LINES_BEFORE_PAPER; i++) {
        _video_run_cpu_for_scanline();
    }
    y += SPECTRUM_BORDER_LINES_BEFORE_PAPER;

    int rowStartAddressAtStartOfThird = 16384;

    // top third of paper
    for (int k = 0; k < SPECTRUM_PAPER_HEIGHT / 3; k++) {
        _video_run_cpu_for_scanline();
    }
    y += SPECTRUM_PAPER_HEIGHT / 3;
    rowStartAddressAtStartOfThird += (SPECTRUM_PAPER_WIDTH / 8) * (SPECTRUM_PAPER_HEIGHT / 3);

    // middle third of paper
    for (int k = 0; k < SPECTRUM_PAPER_HEIGHT / 3; k++) {
        _video_run_cpu_for_scanline();
    }
    y += SPECTRUM_PAPER_HEIGHT / 3;
    rowStartAddressAtStartOfThird += (SPECTRUM_PAPER_WIDTH / 8) * (SPECTRUM_PAPER_HEIGHT / 3);

    // bottom third of paper
    for (int k = 0; k < SPECTRUM_PAPER_HEIGHT / 3; k++) {
        _video_run_cpu_for_scanline();
    }
    y += SPECTRUM_PAPER_HEIGHT / 3;

    ula_bus_go_inactive();

    // border after (below) paper
    for (int i = 0; i < SPECTRUM_BORDER_LINES_AFTER_PAPER; i++) {
        _video_run_cpu_for_scanline();
    }
}

Uint32 _video_callback__via_SDL_Timer(void* param) {
    _video__preframe();
    if (_video_get_must_render_pixels()) {
        _video__frame_with_pixels();
    }
    else {
        _video__frame_without_pixels();
    }
    _video__postframe();

    return timing_get_milliseconds_per_video_frame();
}

void __stdcall _video_callback__via_QueueTimer(PVOID lpParameter, BOOLEAN TimerOrWaitFired) {
    DWORD dwWaitResult = WaitForSingleObject(_videoFrameMutex, INFINITE);
    
    switch (dwWaitResult) {
        case WAIT_OBJECT_0:
            // this thread got ownership of the mutex
            __try {
                _video__preframe();
                if (_video_get_must_render_pixels()) {
                    _video__frame_with_pixels();
                }
                else {
                    _video__frame_without_pixels();
                }
                _video__postframe();
            }

            __finally {
                if (!ReleaseMutex(_videoFrameMutex))
                {
                    // an error happened
                }
            }
            break;

        case WAIT_ABANDONED:
            // this thread got ownership of an abandoned mutex
            return;
    }
}

BOOL video_start__via_SDL_Timer(SDL_Window* window, char zoom, int windowWidth, int windowHeight) {
    log_write("Using a SDL Timer-based video callback");
    _window = window;
    _surface = SDL_GetWindowSurface(window);
    _zoom = zoom;
    _frames = 0;
    _framesAtNormalSpeed = 0;
    _tstates = 0;

    _isDestroying = 0;
    _timerCallbackDestroyed = 0;
    ftime(&_videoStart);

    _surface = SDL_GetWindowSurface(window);
    if (!_surface) {
        log_write("Error: unable to create SDL window surface");
        return FALSE;
    }
    //_video_preload_pixel_cache();

    _entireWindowRectangle.x = 0;
    _entireWindowRectangle.y = 0;
    _entireWindowRectangle.w = _surface->w;
    _entireWindowRectangle.h = _surface->h;

    _renderer = SDL_CreateRenderer(window, -1, SDL_RENDERER_ACCELERATED | SDL_RENDERER_PRESENTVSYNC);
    if (_renderer == NULL) {
        log_write("Error: could not get a video renderer");
        return FALSE;
    }

    _videoIsInitialized = TRUE;

    _timer = SDL_AddTimer(timing_get_milliseconds_per_video_frame(), (SDL_TimerCallback)_video_callback__via_SDL_Timer, NULL);
    log_write("Started timer, yielding CPU control to it");
    return TRUE;
}

BOOL video_start__via_QueueTimer(SDL_Window* window, char zoom, int windowWidth, int windowHeight) {
    log_write("Using a QueueTimer-based video callback");
    _window = window;
    _zoom = zoom;
    _frames = 0;
    _framesAtNormalSpeed = 0;
    _tstates = 0;

    _isDestroying = 0;
    _timerCallbackDestroyed = 0;
    ftime(&_videoStart);

    _surface = SDL_GetWindowSurface(window);
    if (!_surface) {
        log_write("Error: unable to create SDL window surface");
        return FALSE;
    }
    //_video_preload_pixel_cache();

    _entireWindowRectangle.x = 0;
    _entireWindowRectangle.y = 0;
    _entireWindowRectangle.w = _surface->w;
    _entireWindowRectangle.h = _surface->h;

    _renderer = SDL_CreateRenderer(window, -1, SDL_RENDERER_ACCELERATED | SDL_RENDERER_PRESENTVSYNC);
    if (_renderer == NULL) {
        log_write("Error: could not get a video renderer");
        return FALSE;
    }
    
    _videoFrameMutex = CreateMutex(NULL, FALSE, NULL);
    if (!_videoFrameMutex) {
        log_write("Error: unable to create video frame timer mutex");
        return FALSE;
    }
    
    Uint16 framePeriod = timing_get_milliseconds_per_video_frame();
    BOOL res = CreateTimerQueueTimer(&_callbackTimerHandle, NULL, _video_callback__via_QueueTimer, NULL, VIDEO_DELAY_TIMER_START_MS, framePeriod, WT_EXECUTEINTIMERTHREAD);
    if (!res) {
        log_write("Error: unable to start video frame timer");
        _callbackTimerHandle = NULL;
        return FALSE;
    }
    
    _videoIsInitialized = TRUE;
    log_write("Started video timer, yielding CPU control to it");
    return TRUE;
}

SDL_Renderer* video_get_renderer() {
    return _renderer;
}

void _video_report_statistics() {
    char* message = (char*)malloc(500 * sizeof(char));
    if (message == NULL) {
        return;
    }

    double frames = (double)_framesAtNormalSpeed;
    sprintf_s(message, 490, "Total video frames: %lld", (Uint64)frames);
    log_write(message);

    Uint64 runTimeMs = (Uint64)(1000.0 * (_videoEnd.time - _videoStart.time)
        + (_videoEnd.millitm - _videoStart.millitm)) - (Uint32)VIDEO_DELAY_TIMER_START_MS;

    sprintf_s(message, 490, "Total real duration: %I64u ms", runTimeMs);
    log_write(message);

    double fps = frames / ((double)runTimeMs / 1000.f);
    sprintf_s(message, 490, "Average frames per second: %.2f", fps);
    log_write(message);

    double millisPerFrame = (double)runTimeMs / frames;
    Uint32 targetMillisPerFrame = timing_get_milliseconds_per_video_frame();
    sprintf_s(message, 490, "Average milliseconds per frame: %.3f (target is %d)", millisPerFrame, targetMillisPerFrame);
    log_write(message);

    long double tstates = (long double)_tstates;
    long double tstatesPerVideoFrame = tstates / frames;
    Uint64 targetTstatesPerVideoFrame = timing_get_states_per_video_frame();
    long double percentage = (tstatesPerVideoFrame / (long double)targetTstatesPerVideoFrame) * 100.0f;
    sprintf_s(message, 490, "Average tstates per video frame: %.2f (target is %lld), %.1f%%", 
        tstatesPerVideoFrame, 
        targetTstatesPerVideoFrame,
        percentage);
    log_write(message);
    free(message);
}

void video_destroy__via_SDL_Timer() {
    _isDestroying = 1;
    while (!_timerCallbackDestroyed) {}

    SDL_RemoveTimer(_timer);
    ftime(&_videoEnd);

    _video_report_statistics();
}

void video_destroy__via_QueueTimer() {
    if (_callbackTimerHandle) {
        _isDestroying = 1;
        // we busy-wait for callback to become NOOP only if we actually initialized
        // the callback timer successfully
        while (!_timerCallbackDestroyed) {}

        DeleteTimerQueueTimer(NULL, _callbackTimerHandle, NULL);
    }

    ftime(&_videoEnd);

    if (_videoFrameMutex) {
        CloseHandle(_videoFrameMutex);
    }

    if (_surface) {
        SDL_FreeSurface(_surface);
    }

    if (_renderer) {
        SDL_DestroyRenderer(_renderer);
    }

    _video_report_statistics();
}
